# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.


from __future__ import absolute_import
import traceback
import threading
import functools
import json
import time
from obs import const
from obs.client import _BasicClient
from obs.ilog import INFO
from obs.util import jsonLoadsForPy2


def combine(*args):
    return '/'.join(args)


def prepareHeader():
    return {const.CONTENT_TYPE_HEADER: const.MIME_TYPES.get('json')}


def prepareJson(kwargs, pop=None):
    kwargs.pop('self')
    if bool(kwargs):
        if bool(pop):
            for p in list(pop):
                kwargs.pop(p)
        if not bool(kwargs):
            return None
        return json.dumps(kwargs, ensure_ascii=False)
    else:
        return None


def _resultFilter(result, executionState):
    for e in result[::-1]:
        if e.get('execution_state') != executionState:
            result.remove(e)
    return result


def _listWorkflowExecutionMethodName(isJsonResult, defaultMethodName):
    return 'ListWorkflowExecutionResponse' if not isJsonResult else defaultMethodName


def _listWorkflowExecutionCount(isJsonResult, body):
    return body.count if not isJsonResult else body.get('count')


def _listWorkflowExecutionBody(isJsonResult, isPy2, body):
    if isJsonResult:
        body = jsonLoadsForPy2(body) if isPy2 else json.loads(body)
    return body


def _listWorkflowExecutionResult(isJsonResult, body):
    return list(body.executions) if not isJsonResult else list(body.get('executions'))


def _listWorkflowExecutionIsTruncated(isJsonResult, body):
    return body.isTruncated if not isJsonResult else body.get('is_truncated')


def _listWorkflowExecutionNextMarker(isTruncated, isJsonResult, body):
    if isTruncated:
        nextMarker = body.nextMarker if not isJsonResult else body.get('next_marker')
    else:
        nextMarker = None
    return nextMarker


def _listWorkflowExecutionPathArgs(graphName, nextMarker, limit, executionType):
    pathArgs = {'x-workflow-graph-name': graphName}
    if nextMarker:
        pathArgs['x-workflow-next-marker'] = nextMarker
    if limit:
        if limit > 1000:
            raise Exception('Invalid parameter: limit')
        pathArgs['x-workflow-limit'] = limit
    if executionType:
        pathArgs['x-workflow-execution-type'] = executionType
    return pathArgs


def _listWorkflowExecutionResp(isJsonResult, resp, result, isTruncated, nextMarker):
    if isJsonResult:
        resp.body = {
            'count': len(result),
            'is_truncated': isTruncated,
            'executions': result
        }
        if isTruncated:
            resp.body['next_marker'] = nextMarker
        else:
            if resp.body.get('next_marker') is not None:
                del resp.body['next_marker']
        resp.body = json.dumps(resp.body)
    else:
        resp.body.count = len(result)
        resp.body.executions = result
        resp.body.isTruncated = isTruncated
        if isTruncated:
            resp.body.nextMarker = nextMarker
        else:
            if resp.body.nextMarker is not None:
                del resp.body.nextMarker

    return resp


def entrance(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        workflowClient = args[0] if isinstance(args[0], WorkflowClient) else None
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if workflowClient:
                workflowClient.log_client.log(INFO, traceback.format_exc())
            raise e
        finally:
            if workflowClient:
                workflowClient.log_client.log(INFO,
                                              '%s cost %s ms' % (func.__name__, int((time.time() - start) * 1000)))

    return wrapper


class WorkflowClient(_BasicClient):
    _instance_lock = threading.Lock()
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._instance_lock:
                if cls._instance is None:
                    cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self, isJsonResult=False, *args, **kwargs):
        if kwargs.get('path_style') or kwargs.get('is_signature_negotiation'):
            raise Exception('Path style or signature negotiation does not support custom settings for workflow client.')
        super(WorkflowClient, self).__init__(client_mode='workflow', *args, **kwargs)
        self.__resource = 'v2/'
        self.__isJsonResult = isJsonResult
        self.__defaultMethodName = 'GetJsonResponse'

    # begin workflow api
    # begin workflow api
    # begin workflow api

    @entrance
    def createWorkflowTemplate(self, templateName, description=None, states=None, inputs=None, tags=None):
        return self._make_post_request(
            entity=prepareJson(locals(), ['templateName']),
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOW_TEMPLATES, templateName),
            headers=prepareHeader(),
            methodName='CreateWorkflowTemplateResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def getWorkflowTemplate(self, templateName):
        return self._make_get_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOW_TEMPLATES, templateName),
            headers=prepareHeader(),
            methodName='GetWorkflowTemplateResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def deleteWorkflowTemplate(self, templateName):
        return self._make_delete_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOW_TEMPLATES, templateName),
            headers=prepareHeader(),
            methodName=self.__defaultMethodName
        )

    @entrance
    def listWorkflowTemplate(self, templateNamePrefix=None, start=None, limit=None):
        pathArgs = {'x-workflow-prefix': None}
        if start:
            pathArgs['x-workflow-start'] = start
        if limit:
            pathArgs['x-workflow-limit'] = limit

        return self._make_get_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOW_TEMPLATES,
                                                templateNamePrefix if templateNamePrefix is not None else ''),
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName='ListWorkflowTemplateResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def createWorkflow(self, templateName, graphName, agency, description=None, parameters=None):
        return self._make_post_request(
            entity=prepareJson(locals(), ['templateName', 'graphName']),
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOWS, graphName),
            pathArgs={'x-workflow-template-name': templateName},
            headers=prepareHeader(),
            methodName='CreateWorkflowResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def getWorkflow(self, graphName):
        return self._make_get_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOWS, graphName),
            headers=prepareHeader(),
            methodName='GetWorkflowResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def deleteWorkflow(self, graphName):
        return self._make_delete_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOWS, graphName),
            headers=prepareHeader(),
            methodName=self.__defaultMethodName
        )

    @entrance
    def updateWorkflow(self, graphName, parameters=None, description=None):
        return self._make_put_request(
            entity=prepareJson(locals(), ['graphName']),
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOWS, graphName),
            headers=prepareHeader(),
            methodName='UpdateWorkflowResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def listWorkflow(self, graphNamePrefix=None, start=None, limit=None):
        pathArgs = {'x-workflow-prefix': None}
        if start:
            pathArgs['x-workflow-start'] = start
        if limit:
            pathArgs['x-workflow-limit'] = limit

        return self._make_get_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOWS,
                                                graphNamePrefix if graphNamePrefix is not None else ''),
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName='ListWorkflowResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def asyncAPIStartWorkflow(self, graphName, bucket, object, inputs=None):
        return self._make_post_request(
            entity=prepareJson(locals(), ['graphName']),
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOWS, graphName),
            headers=prepareHeader(),
            methodName='AsyncAPIStartWorkflowResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def listWorkflowExecution(self, graphName, executionType=None, nextMarker=None, limit=None, executionState=None):
        pathArgs = _listWorkflowExecutionPathArgs(graphName, nextMarker, limit, executionType)
        if executionState:
            if executionState not in ['RUNNING', 'SUCCESS', 'FAILED']:
                raise Exception('Invalid parameter: execution state')

            resp = self._make_get_request(
                bucketName=None,
                objectKey=self.__resource + const.WORKFLOW_EXECUTIONS,
                pathArgs=pathArgs,
                headers=prepareHeader(),
                methodName=_listWorkflowExecutionMethodName(self.__isJsonResult, self.__defaultMethodName)
            )
            if resp.status > 300:
                return resp

            body = _listWorkflowExecutionBody(self.__isJsonResult, const.IS_PYTHON2, resp.body)

            count = _listWorkflowExecutionCount(self.__isJsonResult, body)

            if count == 0:
                return resp

            result = _listWorkflowExecutionResult(self.__isJsonResult, body)
            isTruncated = _listWorkflowExecutionIsTruncated(self.__isJsonResult, body)
            nextMarker = _listWorkflowExecutionNextMarker(isTruncated, self.__isJsonResult, body)

            result = _resultFilter(result, executionState)

            if len(result) == limit:
                return resp
            else:
                while len(result) < limit and isTruncated:
                    pathArgs['x-workflow-next-marker'] = nextMarker
                    tempResp = self._make_get_request(
                        bucketName=None,
                        objectKey=self.__resource + const.WORKFLOW_EXECUTIONS,
                        pathArgs=pathArgs,
                        headers=prepareHeader(),
                        methodName=_listWorkflowExecutionMethodName(self.__isJsonResult, self.__defaultMethodName)
                    )
                    if tempResp.status > 300:
                        return tempResp

                    body = _listWorkflowExecutionBody(self.__isJsonResult, const.IS_PYTHON2, tempResp.body)

                    count = _listWorkflowExecutionCount(self.__isJsonResult, body)
                    if count == 0:
                        return tempResp

                    tempResult = _listWorkflowExecutionResult(self.__isJsonResult, body)

                    tempResult = _resultFilter(tempResult, executionState)

                    result.extend(tempResult)

                    if len(result) > limit:
                        result = result[0:limit]

                    isTruncated = _listWorkflowExecutionIsTruncated(self.__isJsonResult, body)
                    nextMarker = _listWorkflowExecutionNextMarker(isTruncated, self.__isJsonResult, body)

            return _listWorkflowExecutionResp(self.__isJsonResult, resp, result, isTruncated, nextMarker)
        else:
            return self._make_get_request(
                bucketName=None,
                objectKey=self.__resource + const.WORKFLOW_EXECUTIONS,
                pathArgs=pathArgs,
                headers=prepareHeader(),
                methodName=_listWorkflowExecutionMethodName(self.__isJsonResult, self.__defaultMethodName)
            )

    @entrance
    def getWorkflowExecution(self, executionName, graphName):
        pathArgs = {
            'x-workflow-graph-name': graphName
        }

        return self._make_get_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOW_EXECUTIONS, executionName),
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName='GetWorkflowExecutionResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def restoreFailedWorkflowExecution(self, executionName, graphName):
        pathArgs = {
            'x-workflow-graph-name': graphName
        }

        return self._make_put_request(
            bucketName=None,
            objectKey=self.__resource + combine(const.WORKFLOW_EXECUTIONS, executionName),
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName='RestoreFailedWorkflowExecutionResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def putTriggerPolicy(self, bucketName, rules):
        pathArgs = {
            const.WORKFLOW_TRIGGERPOLICY: None
        }

        return self._make_put_request(
            entity=prepareJson(locals(), ['bucketName']),
            bucketName=bucketName,
            objectKey=None,
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName=self.__defaultMethodName
        )

    @entrance
    def getTriggerPolicy(self, bucketName):
        pathArgs = {
            const.WORKFLOW_TRIGGERPOLICY: None
        }

        return self._make_get_request(
            bucketName=bucketName,
            objectKey=None,
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName='GetTriggerPolicyResponse' if not self.__isJsonResult else self.__defaultMethodName
        )

    @entrance
    def deleteTriggerPolicy(self, bucketName):
        pathArgs = {
            const.WORKFLOW_TRIGGERPOLICY: None
        }

        return self._make_delete_request(
            bucketName=bucketName,
            objectKey=None,
            pathArgs=pathArgs,
            headers=prepareHeader(),
            methodName=self.__defaultMethodName
        )

    # end workflow api
    # end workflow api
    # end workflow api



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.

import threading

LOCK_COUNT = 16
lock_list = []

for i in range(LOCK_COUNT):
    lock_list.append(threading.RLock())


def get_lock(index):
    if index < 0 or index >= LOCK_COUNT:
        raise Exception('cannot find a valid lock')

    return lock_list[index]



#!/usr/bin/python
# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import print_function
import hyper
from hyper.common import headers
from hyper.common import util
from hyper.http20 import stream
from hyper.common import exceptions
from obs import const
import threading


def _is_hyper_exception(e):
    return isinstance(e, (exceptions.SocketError, exceptions.InvalidResponseError, exceptions.ConnectionResetError))


def _get_server_connection(server, port=None, context=None, is_secure=False, proxy_host=None, proxy_port=None):
    try:
        return HTTP20ConnectionWrapper(host=server, port=port, ssl_context=context, secure=is_secure,
                                       proxy_host=proxy_host, proxy_port=proxy_port)
    except Exception as e:
        raise e


def to_string(item):
    try:
        return str(item) if item is not None else ''
    except Exception:
        return ''


def _get_ssl_context(ssl_verify):
    try:
        from hyper import tls
        import ssl
        import os
        context = tls.init_context(None, None, None)
        context.check_hostname = False
        if ssl_verify:
            cafile = to_string(ssl_verify)
            if os.path.isfile(cafile):
                context.load_verify_locations(cafile)
            context.verify_mode = ssl.CERT_REQUIRED
        else:
            context.verify_mode = ssl.CERT_NONE
        return context
    except Exception:
        import traceback
        print(traceback.format_exc())


def canonical_form(k, v):
    yield k, v


headers.canonical_form = canonical_form


def _send_chunk(self, data, final):
    while len(data) > self._out_flow_control_window:
        self._recv_cb()

    end_stream = False
    if len(data) <= stream.MAX_CHUNK and final:
        end_stream = True

    with self._conn as conn:
        conn.send_data(
            stream_id=self.stream_id, data=data, end_stream=end_stream
        )
    self._send_outstanding_data()

    if end_stream:
        self.local_closed = True


def send_data(self, data, final):
    chunks = [data[i:i + stream.MAX_CHUNK]
              for i in range(0, len(data), stream.MAX_CHUNK)]

    index = 0
    count = len(chunks)
    for chunk in chunks:
        self._send_chunk(chunk, final and index == count - 1)
        index += 1


stream.Stream._send_chunk = _send_chunk
stream.Stream.send_data = send_data


class HTTP20ConnectionWrapper(hyper.HTTP20Connection):

    def __init__(self, host, port=None, secure=None, window_manager=None,
                 enable_push=False, ssl_context=None, proxy_host=None,
                 proxy_port=None, force_proto=None, **kwargs):
        self._stream_id_context = threading.local()
        super(HTTP20ConnectionWrapper, self).__init__(host, port=port, secure=secure,
                                                      window_manager=window_manager, enable_push=enable_push,
                                                      ssl_context=ssl_context, proxy_host=proxy_host,
                                                      proxy_port=proxy_port, force_proto=force_proto, **kwargs)

    def getresponse(self, stream_id=None):
        stream_id = stream_id or (
            self._stream_id_context.stream_id if hasattr(self._stream_id_context, 'stream_id') else None)
        stream = self._get_stream(stream_id)
        return HTTP20ResponseWrapper(stream.getheaders(), stream)

    def request(self, method, url, body=None, headers=None):
        headers = headers or {}
        if const.HOST_HEADER in headers:
            headers[':authority'] = headers[const.HOST_HEADER]

        with self._write_lock:
            stream_id = self.putrequest(method, url)
            self._stream_id_context.stream_id = stream_id
            default_headers = (':method', ':scheme', ':authority', ':path')
            for name, value in headers.items():
                is_default = util.to_native_string(name) in default_headers
                if isinstance(value, list):
                    for item in value:
                        self.putheader(name, item, stream_id, replace=is_default)
                else:
                    self.putheader(name, value, stream_id, replace=is_default)

            final = True
            message_body = body
            if body is not None:
                if callable(body):
                    final = False
                    message_body = None
                else:
                    if isinstance(body, (unicode, bytes)):
                        message_body = util.to_bytestring(body)
                    if len(message_body) == 0:
                        message_body = None

            self.endheaders(message_body=message_body, final=final, stream_id=stream_id)
            if not final:
                body(self)

            return stream_id

    def send(self, data, final=False, stream_id=None):
        stream_id = stream_id or (
            self._stream_id_context.stream_id if hasattr(self._stream_id_context, 'stream_id') else None)
        stream = self._get_stream(stream_id)
        stream.send_data(data, final)


class HTTP20ResponseWrapper(hyper.HTTP20Response):
    def getheader(self, key, default_value=None):
        ret = self.headers.get(key)
        return ret if ret is not None else default_value

    def getheaders(self):
        return self.headers



#!/usr/bin/python
# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.

import sys
import platform
import os

READ_ONCE_LENGTH = 65536

CONTENT_LENGTH_HEADER = 'Content-Length'
CONTENT_TYPE_HEADER = 'Content-Type'
CONTENT_MD5_HEADER = 'Content-MD5'
CACHE_CONTROL_HEADER = 'Cache-Control'
CONTENT_DISPOSITION_HEADER = 'Content-Disposition'
CONTENT_ENCODING_HEADER = 'Content-Encoding'
CONTENT_LANGUAGE_HEADER = 'Content-Language'
EXPIRES_HEADER = 'Expires'

DATE_HEADER = 'Date'

CONTENT_LIST = [CONTENT_TYPE_HEADER.lower(), CONTENT_MD5_HEADER.lower(), DATE_HEADER.lower()]

HOST_HEADER = 'Host'
AUTHORIZATION_HEADER = 'Authorization'
CONNECTION_HEADER = 'Connection'
PROXY_CONNECTION_HEADER = 'Proxy-Connection'
PROXY_AUTHORIZATION_HEADER = 'Proxy-Authorization'
USER_AGENT_HEADER = 'User-Agent'
TRANSFER_ENCODING_HEADER = 'Transfer-Encoding'
TRANSFER_ENCODING_VALUE = 'chunked'
ORIGIN_HEADER = 'Origin'
RANGE_HEADER = 'Range'
IF_MODIFIED_SINCE = 'If-Modified-Since'
IF_UNMODIFIED_SINCE = 'If-Unmodified-Since'
IF_MATCH = 'If-Match'
IF_NONE_MATCH = 'If-None-Match'

CONNECTION_KEEP_ALIVE_VALUE = 'Keep-Alive'
CONNECTION_CLOSE_VALUE = 'close'
ACCESS_CONTROL_REQUEST_METHOD_HEADER = 'Access-Control-Request-Method'
ACCESS_CONTROL_REQUEST_HEADERS_HEADER = 'Access-Control-Request-Headers'
LOCATION_HEADER = 'Location'
ETAG_HEADER = 'ETag'
LAST_MODIFIED_HEADER = 'Last-Modified'

VERSION_ID_PARAM = 'versionId'
RESPONSE_CACHE_CONTROL_PARAM = 'response-cache-control'
RESPONSE_CONTENT_DISPOSITION_PARAM = 'response-content-disposition'
RESPONSE_CONTENT_ENCODING_PARAM = 'response-content-encoding'
RESPONSE_CONTENT_LANGUAGE_PARAM = 'response-content-language'
RESPONSE_CONTENT_TYPE_PARAM = 'response-content-type'
RESPONSE_EXPIRES_PARAM = 'response-expires'
X_IMAGE_PROCESS_PARAM = 'x-image-process'

HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_POST = 'POST'
HTTP_METHOD_GET = 'GET'
HTTP_METHOD_DELETE = 'DELETE'
HTTP_METHOD_HEAD = 'HEAD'
HTTP_METHOD_OPTIONS = 'OPTIONS'

IS_WINDOWS = platform.system() == 'Windows' or os.name == 'nt'
IS_PYTHON2 = sys.version_info.major == 2 or sys.version < '3'
IS_PYTHON35_UP = sys.version >= '3.5'
BASESTRING = basestring if IS_PYTHON2 else str
UNICODE = unicode if IS_PYTHON2 else str
LONG = long if IS_PYTHON2 else int

DEFAULT_SECURE_PORT = 443
DEFAULT_INSECURE_PORT = 80
DEFAULT_MINIMUM_SIZE = 100 * 1024
DEFAULT_MAXIMUM_SIZE = 5 * 1024 * 1024 * 1024

DEFAULT_BYTE_INTTERVAL = 102400
DEFAULT_TASK_INTTERVAL = 100
DEFAULT_TASK_NUM = 8
DEFAULT_TASK_QUEUE_SIZE = 20000

OBS_SDK_VERSION = '3.21.8'

V2_META_HEADER_PREFIX = 'x-amz-meta-'
V2_HEADER_PREFIX = 'x-amz-'

OBS_META_HEADER_PREFIX = 'x-obs-meta-'
OBS_HEADER_PREFIX = 'x-obs-'

GMT_DATE_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'
LONG_DATE_FORMAT = '%Y%m%dT%H%M%SZ'
SHORT_DATE_FORMAT = '%Y%m%d'
EXPIRATION_DATE_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

CANONICAL_REQUEST = 'CannonicalRequest'
CANONICAL_STRING = 'CanonicalString'

IPv4_REGEX = '^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$'

V2_SIGNATURE = 'v2'
OBS_SIGNATURE = 'obs'

FETCH_POLICY_KEY = "v1/extension_policy"
FETCH_JOB_KEY = "v1/async-fetch/jobs"

WORKFLOW_TEMPLATES = 'workflowtemplates'
WORKFLOWS = 'workflows'
WORKFLOW_EXECUTIONS = 'workflowexecutions'
WORKFLOW_TRIGGERPOLICY = 'obsworkflowtriggerpolicy'

ALLOWED_RESOURCE_PARAMTER_NAMES = (
    'acl',
    'backtosource',
    'policy',
    'torrent',
    'logging',
    'location',
    'storageinfo',
    'quota',
    'storageclass',
    'storagepolicy',
    'requestpayment',
    'versions',
    'versioning',
    'versionid',
    'uploads',
    'uploadid',
    'partnumber',
    'website',
    'notification',
    'lifecycle',
    'deletebucket',
    'delete',
    'cors',
    'restore',
    'tagging',
    'replication',
    'metadata',
    'encryption',

    # File System API
    'append',
    'position',
    'truncate',
    'modify',
    'rename',
    'length',
    'name',
    'fileinterface',

    'response-content-type',
    'response-content-language',
    'response-expires',
    'response-cache-control',
    'response-content-disposition',
    'response-content-encoding',
    'x-image-save-bucket',
    'x-image-save-object',
    'x-image-process',
    'x-oss-process',

    # workflow api
    'x-workflow-prefix',
    'x-workflow-start',
    'x-workflow-limit',
    'x-workflow-template-name',
    'x-workflow-graph-name',
    'x-workflow-execution-state',
    'x-workflow-execution-type',
    'x-workflow-next-marker',
    'obsworkflowtriggerpolicy'
)

ALLOWED_REQUEST_HTTP_HEADER_METADATA_NAMES = (
    'content-type',
    'content-md5',
    'content-length',
    'content-language',
    'expires',
    'origin',
    'cache-control',
    'content-disposition',
    'content-encoding',
    'access-control-request-method',
    'access-control-request-headers',
    'success-action-redirect',
    'x-default-storage-class',
    'location',
    'date',
    'etag',
    'range',
    'host',
    'if-modified-since',
    'if-unmodified-since',
    'if-match',
    'if-none-match',
    'last-modified',
    'content-range'
)

ALLOWED_RESPONSE_HTTP_HEADER_METADATA_NAMES = (
    'content-type',
    'content-md5',
    'content-length',
    'content-language',
    'expires',
    'origin',
    'cache-control',
    'content-disposition',
    'content-encoding',
    'x-default-storage-class',
    'location',
    'date',
    'etag',
    'host',
    'last-modified',
    'content-range',
    'x-reserved',
    'access-control-allow-origin',
    'access-control-allow-headers',
    'access-control-max-age',
    'access-control-allow-methods',
    'access-control-expose-headers',
    'connection',
    'x-reserved-indicator',
    'x-oef-request-id'
)

MIME_TYPES = {
    '7z': 'application/x-7z-compressed',
    'aac': 'audio/x-aac',
    'ai': 'application/postscript',
    'aif': 'audio/x-aiff',
    'apk': 'application/vnd.android.package-archive',
    'asc': 'text/plain',
    'asf': 'video/x-ms-asf',
    'atom': 'application/atom+xml',
    'avi': 'video/x-msvideo',
    'bmp': 'image/bmp',
    'bz2': 'application/x-bzip2',
    'cer': 'application/pkix-cert',
    'crl': 'application/pkix-crl',
    'crt': 'application/x-x509-ca-cert',
    'css': 'text/css',
    'csv': 'text/csv',
    'cu': 'application/cu-seeme',
    'deb': 'application/x-debian-package',
    'doc': 'application/msword',
    'docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    'dvi': 'application/x-dvi',
    'eot': 'application/vnd.ms-fontobject',
    'eps': 'application/postscript',
    'epub': 'application/epub+zip',
    'etx': 'text/x-setext',
    'flac': 'audio/flac',
    'flv': 'video/x-flv',
    'gif': 'image/gif',
    'gz': 'application/gzip',
    'htm': 'text/html',
    'html': 'text/html',
    'ico': 'image/x-icon',
    'ics': 'text/calendar',
    'ini': 'text/plain',
    'iso': 'application/x-iso9660-image',
    'jar': 'application/java-archive',
    'jpe': 'image/jpeg',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpeg',
    'js': 'text/javascript',
    'json': 'application/json',
    'latex': 'application/x-latex',
    'log': 'text/plain',
    'm4a': 'audio/mp4',
    'm4v': 'video/mp4',
    'mid': 'audio/midi',
    'midi': 'audio/midi',
    'mov': 'video/quicktime',
    'mp3': 'audio/mpeg',
    'mp4': 'video/mp4',
    'mp4a': 'audio/mp4',
    'mp4v': 'video/mp4',
    'mpe': 'video/mpeg',
    'mpeg': 'video/mpeg',
    'mpg': 'video/mpeg',
    'mpg4': 'video/mp4',
    'oga': 'audio/ogg',
    'ogg': 'audio/ogg',
    'ogv': 'video/ogg',
    'ogx': 'application/ogg',
    'pbm': 'image/x-portable-bitmap',
    'pdf': 'application/pdf',
    'pgm': 'image/x-portable-graymap',
    'png': 'image/png',
    'pnm': 'image/x-portable-anymap',
    'ppm': 'image/x-portable-pixmap',
    'ppt': 'application/vnd.ms-powerpoint',
    'pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    'ps': 'application/postscript',
    'qt': 'video/quicktime',
    'rar': 'application/x-rar-compressed',
    'ras': 'image/x-cmu-raster',
    'rss': 'application/rss+xml',
    'rtf': 'application/rtf',
    'sgm': 'text/sgml',
    'sgml': 'text/sgml',
    'svg': 'image/svg+xml',
    'swf': 'application/x-shockwave-flash',
    'tar': 'application/x-tar',
    'tif': 'image/tiff',
    'tiff': 'image/tiff',
    'torrent': 'application/x-bittorrent',
    'ttf': 'application/x-font-ttf',
    'txt': 'text/plain',
    'wav': 'audio/x-wav',
    'webm': 'video/webm',
    'wma': 'audio/x-ms-wma',
    'wmv': 'video/x-ms-wmv',
    'woff': 'application/x-font-woff',
    'wsdl': 'application/wsdl+xml',
    'xbm': 'image/x-xbitmap',
    'xls': 'application/vnd.ms-excel',
    'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    'xml': 'application/xml',
    'xpm': 'image/x-xpixmap',
    'xwd': 'image/x-xwindowdump',
    'yaml': 'text/yaml',
    'yml': 'text/yaml',
    'zip': 'application/zip'
}



#!/usr/bin/python
# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.

import base64
import hashlib
import json
import re

from obs import const, progress

if const.IS_PYTHON2:
    import urllib
else:
    import urllib.parse as urllib
from obs.ilog import INFO, ERROR


def to_bool(item):
    try:
        return True if item is not None and str(item).lower() == 'true' else False
    except Exception:
        return None


def to_int(item):
    try:
        return int(item)
    except Exception:
        return None


def to_long(item):
    try:
        return const.LONG(item)
    except Exception:
        return None


def to_float(item):
    try:
        return float(item)
    except Exception:
        return None


def to_string(item):
    try:
        return str(item) if item is not None else ''
    except Exception:
        return ''


def is_valid(item):
    return item is not None and item.strip() != ''


class RequestFormat(object):

    @staticmethod
    def get_path_format():
        return PathFormat()

    @staticmethod
    def get_sub_domain_format():
        return SubdomainFormat()

    @classmethod
    def convert_path_string(cls, path_args, allowdNames=None):
        e = ''
        if isinstance(path_args, dict):
            e1 = '?'
            e2 = '&'
            for path_key, path_value in path_args.items():
                flag = True
                if allowdNames is not None and path_key not in allowdNames:
                    flag = False
                if flag:
                    path_key = encode_item(path_key, '/')
                    if path_value is None:
                        e1 += path_key + '&'
                        continue
                    e2 += path_key + '=' + encode_item(path_value, '/') + '&'
            e = (e1 + e2).replace('&&', '&').replace('?&', '?')[:-1]
        return e

    def get_endpoint(self, server, port, bucket):
        return

    def get_path_base(self, bucket, key):
        return

    def get_url(self, bucket, key, path_args):
        return


class PathFormat(RequestFormat):

    @staticmethod
    def get_server(server, bucket):
        return server

    def get_path_base(self, bucket, key):
        if bucket:
            return '/' + bucket + '/' + encode_object_key(key) if key else '/' + bucket
        return '/' + encode_object_key(key) if key else '/'

    def get_endpoint(self, server, port, bucket):
        if port == 80 or port == 443:
            return server
        return server + ':' + str(port)

    def get_url(self, bucket, key, path_args):
        path_base = self.get_path_base(bucket, key)
        path_arguments = self.convert_path_string(path_args)
        return path_base + path_arguments

    def get_full_url(self, is_secure, server, port, bucket, key, path_args):
        url = 'https://' if is_secure else 'http://'
        url += self.get_endpoint(server, port, bucket)
        url += self.get_url(bucket, key, path_args)
        return url


class SubdomainFormat(RequestFormat):

    @staticmethod
    def get_server(server, bucket):
        return bucket + '.' + server if bucket else server

    def get_path_base(self, bucket, key):
        if key is None:
            return '/'
        return '/' + encode_object_key(key)

    def get_endpoint(self, server, port, bucket):
        if port == 80 or port == 443:
            return self.get_server(server, bucket)
        return self.get_server(server, bucket) + ':' + str(port)

    def get_url(self, bucket, key, path_args):
        url = self.convert_path_string(path_args)
        return self.get_path_base(bucket, key) + url

    def get_full_url(self, is_secure, server, port, bucket, key, path_args):
        url = 'https://' if is_secure else 'http://'
        url += self.get_endpoint(server, port, bucket)
        url += self.get_url(bucket, key, path_args)
        return url


class delegate(object):
    def __init__(self, conn):
        self.conn = conn

    def send(self, data, final=False, stream_id=None):
        self.conn.send(data)


def conn_delegate(conn):
    return delegate(conn)


def get_readable_entity(readable, chunk_size=const.READ_ONCE_LENGTH, notifier=None, auto_close=True):
    if notifier is None:
        notifier = progress.NONE_NOTIFIER

    def entity(conn):
        try:
            while True:
                chunk = readable.read(chunk_size)
                if not chunk:
                    conn.send('0\r\n\r\n' if const.IS_PYTHON2 else '0\r\n\r\n'.encode('UTF-8'), final=True)
                    break
                newReadCount = len(chunk)
                if newReadCount > 0:
                    notifier.send(newReadCount)
                hex_chunk = hex(len(chunk))[2:]
                conn.send(hex_chunk if const.IS_PYTHON2 else hex_chunk.encode('UTF-8'))
                conn.send('\r\n' if const.IS_PYTHON2 else '\r\n'.encode('UTF-8'))
                conn.send(chunk)
                conn.send('\r\n' if const.IS_PYTHON2 else '\r\n'.encode('UTF-8'))
        finally:
            if hasattr(readable, 'close') and callable(readable.close) and auto_close:
                readable.close()

    return entity


def get_readable_entity_by_total_count(readable, totalCount, chunk_size=const.READ_ONCE_LENGTH, notifier=None,
                                       auto_close=True):
    return get_entity_for_send_with_total_count(readable, totalCount, chunk_size, notifier, auto_close)


def get_file_entity_by_total_count(file_path, totalCount, chunk_size=const.READ_ONCE_LENGTH, notifier=None):
    f = open(file_path, "rb")
    return get_entity_for_send_with_total_count(f, totalCount, chunk_size, notifier)


def get_entity_for_send_with_total_count(readable, totalCount=None, chunk_size=const.READ_ONCE_LENGTH, notifier=None,
                                         auto_close=True):
    if notifier is None:
        notifier = progress.NONE_NOTIFIER

    def entity(conn):
        readCount = 0
        try:
            while True:
                if totalCount is None or totalCount - readCount >= chunk_size:
                    readCountOnce = chunk_size
                else:
                    readCountOnce = totalCount - readCount
                chunk = readable.read(readCountOnce)
                newReadCount = len(chunk)
                if newReadCount > 0:
                    notifier.send(newReadCount)
                readCount += newReadCount
                if (totalCount is not None and readCount >= totalCount) or (totalCount is not None and not chunk):
                    conn.send(chunk, final=True)
                    break
                conn.send(chunk)
        finally:
            if hasattr(readable, 'close') and callable(readable.close) and auto_close:
                readable.close()

    return entity


def get_file_entity_by_offset_partsize(file_path, offset, totalCount, chunk_size=const.READ_ONCE_LENGTH, notifier=None):
    f = open(file_path, "rb")
    f.seek(offset)
    return get_entity_for_send_with_total_count(f, totalCount, chunk_size, notifier)


def is_ipaddress(item):
    return re.match(const.IPv4_REGEX, item)


def md5_encode(unencoded):
    m = hashlib.md5()
    unencoded = unencoded if const.IS_PYTHON2 else (
        unencoded.encode('UTF-8') if not isinstance(unencoded, bytes) else unencoded)
    m.update(unencoded)
    return m.digest()


def covert_string_to_bytes(str_object):
    if not const.IS_PYTHON2:
        if isinstance(str_object, str):
            return str_object.encode("UTF-8")
    return str_object


def base64_encode(unencoded):
    unencoded = unencoded if const.IS_PYTHON2 else (
        unencoded.encode('UTF-8') if not isinstance(unencoded, bytes) else unencoded)
    encode_str = base64.b64encode(unencoded, altchars=None)
    return encode_str if const.IS_PYTHON2 else encode_str.decode('UTF-8')


def encode_object_key(key):
    return encode_item(key, '/~')


def encode_item(item, safe='/'):
    return urllib.quote(to_string(item), safe)


def decode_item(item):
    return urllib.unquote(item)


def safe_trans_to_utf8(item):
    if not const.IS_PYTHON2:
        return item
    if item is not None:
        item = safe_encode(item)
        try:
            return item.decode('GB2312').encode('UTF-8')
        except Exception:
            return item
    return None


def safe_trans_to_gb2312(item):
    if not const.IS_PYTHON2:
        return item
    if item is not None:
        item = safe_encode(item)
        try:
            return item.decode('UTF-8').encode('GB2312')
        except Exception:
            return item
    return None


def safe_decode(item):
    if not const.IS_PYTHON2:
        return item
    if isinstance(item, str):
        try:
            item = item.decode('UTF-8')
        except Exception:
            try:
                item = item.decode('GB2312')
            except Exception:
                item = None
    return item


def safe_encode(item):
    if not const.IS_PYTHON2:
        return item
    if isinstance(item, const.UNICODE):
        try:
            item = item.encode('UTF-8')
        except UnicodeDecodeError:
            try:
                item = item.encode('GB2312')
            except Exception:
                item = None
    return item


def md5_file_encode_by_size_offset(file_path=None, size=None, offset=None, chuckSize=None):
    if file_path is not None and size is not None and offset is not None:
        m = hashlib.md5()
        with open(file_path, 'rb') as fp:
            CHUNK_SIZE = const.READ_ONCE_LENGTH if chuckSize is None else chuckSize
            fp.seek(offset)
            read_count = 0
            while read_count < size:
                read_size = CHUNK_SIZE if size - read_count >= CHUNK_SIZE else size - read_count
                data = fp.read(read_size)
                read_count_once = len(data)
                if read_count_once <= 0:
                    break
                m.update(data)
                read_count += read_count_once
        return m.digest()


def do_close(result, conn, connHolder, log_client=None):
    if not result:
        close_conn(conn, log_client)
    elif result.getheader('connection', '').lower() == 'close':
        if log_client:
            log_client.log(INFO, 'server inform to close connection')
        close_conn(conn, log_client)
    elif to_int(result.status) >= 500 or connHolder is None:
        close_conn(conn, log_client)
    elif hasattr(conn, '_clear') and conn._clear:
        close_conn(conn, log_client)
    else:
        if connHolder is not None:
            try:
                connHolder['connSet'].put_nowait(conn)
            except Exception:
                close_conn(conn, log_client)


def close_conn(conn, log_client=None):
    try:
        if conn:
            conn.close()
    except Exception as ex:
        if log_client:
            log_client.log(ERROR, ex)


SKIP_VERIFY_ATTR_TYPE = False


def verify_attr_type(value, allowedAttrType):
    if SKIP_VERIFY_ATTR_TYPE:
        return True
    if isinstance(allowedAttrType, list):
        for t in allowedAttrType:
            if isinstance(value, t):
                return True
        return False
    return isinstance(value, allowedAttrType)


def lazyCallback(*args, **kwargs):
    pass


def jsonLoadsForPy2(json_text):
    return _byteify(json.loads(json_text, object_hook=_byteify), ignore_dicts=True)


def _byteify(data, ignore_dicts=False):
    if isinstance(data, const.UNICODE):
        return data.encode('utf-8')
    if isinstance(data, list):
        return [_byteify(item, ignore_dicts=True) for item in data]
    if isinstance(data, dict) and not ignore_dicts:
        return {
            _byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True)
            for key, value in data.iteritems()
        }
    return data



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.

import threading
from obs import const
from obs import util

if const.IS_PYTHON2:
    import Queue as queue
else:
    import queue


class ThreadPool(object):

    def __init__(self, thread_size=const.DEFAULT_TASK_NUM, queue_size=const.DEFAULT_TASK_QUEUE_SIZE):
        self.thread_size = thread_size
        self._alive_threads = 0
        self._task_queue = queue.Queue(queue_size)
        self._threads = []
        self._init_threads()
        self._shutdown_lock = threading.Lock()

    def _init_threads(self):
        for i in range(self.thread_size):
            self._alive_threads += 1
            work_thread = threading.Thread(target=self._run)
            self._threads.append(work_thread)
            work_thread.start()

    def _run(self):
        task = self._task_queue.get()
        while task is not None:
            (func, args, kwargs, future) = task

            if future is None:
                result = func(*args, **kwargs)
            else:
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    future.set_exception(e)
                else:
                    future.set_result(result)

            del task

            task = self._task_queue.get()

    def execute(self, func, *args, **kwargs):
        task = (func, args, kwargs, None)
        self._task_queue.put(task)

    def submit(self, func, *args, **kwargs):
        future = Future()
        task = (func, args, kwargs, future)
        self._task_queue.put(task)
        return future

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            while self._alive_threads:
                self._task_queue.put(None)
                self._alive_threads -= 1
            if wait:
                for t in self._threads:
                    t.join()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False


class TimeoutError(Exception):
    pass


PENDING = 'PENDING'
COMPLETED = 'COMPLETED'


class Future(object):
    def __init__(self):
        self._condition = threading.Condition()
        self._state = PENDING
        self._result = None
        self._exception = None
        self._callback = None

    def set_result(self, result):
        with self._condition:
            self._result = result
            self._state = COMPLETED
            self._condition.notify_all()

        if self._callback:
            self._callback(self)

    def set_exception(self, exception):
        with self._condition:
            self._exception = exception
            self._state = COMPLETED
            self._condition.notify_all()

        if self._callback:
            self._callback(self)

    def set_callback(self, callback):
        with self._condition:
            if self._state is PENDING:
                self._callback = callback
                return
        callback(self)

    def _get_result(self):
        if self._exception:
            raise self._exception  # pylint: disable=raising-bad-type
        else:
            return self._result

    def get_result(self, timeout=None):
        with self._condition:
            if self._state == COMPLETED:
                return self._get_result()

            self._condition.wait(timeout)

            if self._state == COMPLETED:
                return self._get_result()
            else:
                raise TimeoutError()

    def get_exception(self, timeout=None):
        with self._condition:
            if self._state == COMPLETED:
                return self._exception

            self._condition.wait(timeout)

            if self._state == COMPLETED:
                return self._exception
            else:
                raise TimeoutError()


class ExecuteProgress(object):
    def __init__(self):
        self.successful_tasks = 0
        self._successful_lock = threading.Lock()
        self.failed_tasks = 0
        self._failed_lock = threading.Lock()
        self.finished_tasks = 0
        self._finished_lock = threading.Lock()
        self.total_tasks = 0

    def _successful_increment(self):
        with self._successful_lock:
            self.successful_tasks += 1
            return self.successful_tasks

    def _failed_increment(self):
        with self._failed_lock:
            self.failed_tasks += 1
            return self.failed_tasks

    def _finished_increment(self):
        with self._finished_lock:
            self.finished_tasks += 1
            return self.finished_tasks

    def get_successful_tasks(self):
        with self._successful_lock:
            return self.successful_tasks

    def get_failed_tasks(self):
        with self._failed_lock:
            return self.failed_tasks

    def get_finished_tasks(self):
        with self._finished_lock:
            return self.finished_tasks

    def get_total_tasks(self):
        return self.total_tasks


def _reportProgress(progress, interval, progressCallback):
    finishedTasks = progress._finished_increment()
    if finishedTasks % interval == 0 or finishedTasks == progress.get_total_tasks():
        successfulTasks = progress.get_successful_tasks()
        failedTasks = progress.get_failed_tasks()
        progressCallback(successfulTasks, failedTasks, progress.get_total_tasks())


def _checkBulkTasksPara(task_num, task_queue_size, task_interval, threshold):
    origine = [task_num, task_queue_size, task_interval, threshold]
    default = (
        const.DEFAULT_TASK_NUM, const.DEFAULT_TASK_QUEUE_SIZE, const.DEFAULT_BYTE_INTTERVAL, const.DEFAULT_MAXIMUM_SIZE)
    size = len(origine)
    for i in range(size):
        origine[i] = util.to_int(origine[i])
        if origine[i] is None or origine[i] <= 0:
            origine[i] = default[i]
    return tuple(origine)



#!/usr/bin/python
# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

import functools
import json
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
import math
import operator
import os
import sys
import threading
import traceback

from obs import const, progress, util
from obs.ilog import DEBUG, ERROR, INFO
from obs.model import BaseModel, CompleteMultipartUploadRequest, CompletePart, GetObjectHeader, GetObjectRequest, \
    UploadFileHeader

if const.IS_PYTHON2:
    import Queue as queue
else:
    import queue


def _resume_upload(bucketName, objectKey, uploadFile, partSize, taskNum, enableCheckPoint, checkPointFile, checkSum,
                   metadata, progressCallback, obsClient, headers, extensionHeaders=None, encoding_type=None):
    upload_operation = uploadOperation(util.to_string(bucketName), util.to_string(objectKey),
                                       util.to_string(uploadFile), partSize, taskNum, enableCheckPoint,
                                       util.to_string(checkPointFile), checkSum, metadata, progressCallback, obsClient,
                                       headers, extensionHeaders=extensionHeaders, encoding_type=encoding_type)
    return upload_operation._upload()


def _resume_download(bucketName, objectKey, downloadFile, partSize, taskNum, enableCheckPoint, checkPointFile,
                     header, versionId, progressCallback, obsClient, imageProcess=None,
                     notifier=progress.NONE_NOTIFIER, extensionHeaders=None):
    down_operation = downloadOperation(util.to_string(bucketName), util.to_string(objectKey),
                                       util.to_string(downloadFile), partSize, taskNum, enableCheckPoint,
                                       util.to_string(checkPointFile),
                                       header, versionId, progressCallback, obsClient, imageProcess, notifier,
                                       extensionHeaders=extensionHeaders)

    return _resume_download_with_operation(down_operation)


def _resume_download_with_operation(down_operation):
    if down_operation.size == 0:
        down_operation._delete_record()
        down_operation._delete_tmp_file()
        with open(down_operation.fileName, 'wb'):
            pass
        if down_operation.progressCallback is not None and callable(down_operation.progressCallback):
            down_operation.progressCallback(0, 0, 0)
        return down_operation._metadata_resp
    return down_operation._download()


class Operation(object):
    def __init__(self, bucketName, objectKey, fileName, partSize, taskNum, enableCheckPoint, checkPointFile,
                 progressCallback, obsClient, notifier=progress.NONE_NOTIFIER):
        self.bucketName = bucketName
        self.objectKey = objectKey
        self.fileName = util.safe_trans_to_gb2312(fileName)
        self.partSize = partSize
        self.taskNum = taskNum
        self.enableCheckPoint = enableCheckPoint
        self.checkPointFile = util.safe_trans_to_gb2312(checkPointFile)
        self.progressCallback = progressCallback
        self.notifier = notifier
        self.obsClient = obsClient
        self._lock = threading.Lock()
        self._abortLock = threading.Lock()
        self._abort = False
        self._record = None
        self._exception = None

    def _is_abort(self):
        with self._abortLock:
            return self._abort

    def _do_abort(self, error):
        with self._abortLock:
            self._abort = True
            if self._exception is None:
                self._exception = error

    def _get_record(self):
        self.obsClient.log_client.log(INFO, 'load checkpoint file...')
        if os.path.exists(self.checkPointFile):
            try:
                with open(self.checkPointFile, 'r') as f:
                    content = json.load(f)
                    return _parse_string(content)
            except ValueError:
                pass
        return None

    def _delete_record(self):
        if os.path.exists(self.checkPointFile):
            os.remove(self.checkPointFile)
            self.obsClient.log_client.log(INFO,
                                          'delete checkpoint file success. path is:{0}'.format(self.checkPointFile))

    def _write_record(self, record):
        with open(self.checkPointFile, 'w') as f:
            json.dump(record, f)
            self.obsClient.log_client.log(INFO,
                                          'write checkpoint file success. file path is {0}'.format(self.checkPointFile))


class uploadOperation(Operation):
    def __init__(self, bucketName, objectKey, uploadFile, partSize, taskNum, enableCheckPoint, checkPointFile,
                 checkSum, metadata, progressCallback, obsClient, headers, extensionHeaders=None, encoding_type=None):
        if enableCheckPoint and not checkPointFile:
            checkPointFile = uploadFile + '.upload_record'
        super(uploadOperation, self).__init__(bucketName, objectKey, uploadFile, partSize, taskNum, enableCheckPoint,
                                              checkPointFile, progressCallback, obsClient)
        self.checkSum = checkSum
        self.metadata = metadata
        self.headers = headers
        self.extensionHeaders = extensionHeaders
        self.encoding_type = encoding_type

        try:
            self.size = os.path.getsize(self.fileName)
            self.lastModified = os.path.getmtime(self.fileName)
        except Exception as e:
            self._delete_record()
            self.obsClient.log_client.log(ERROR,
                                          'something is happened when obtain uploadFile information. Please check')
            raise e
        self.total_parts = 0
        self.reset_partSize()
        self._record = {'bucketName': self.bucketName, 'objectKey': self.objectKey, 'uploadFile': self.fileName,
                        'partEtags': []}
        resp = self.obsClient.headBucket(self.bucketName, extensionHeaders=extensionHeaders)
        if resp.status > 300:
            raise Exception('head bucket {0} failed. Please check. Status:{1}.'.format(self.bucketName, resp.status))

    def reset_partSize(self):
        if self.partSize < const.DEFAULT_MINIMUM_SIZE:
            self.partSize = const.DEFAULT_MINIMUM_SIZE
        elif self.partSize > const.DEFAULT_MAXIMUM_SIZE:
            self.partSize = const.DEFAULT_MAXIMUM_SIZE
        else:
            self.partSize = util.to_int(self.partSize)
        if self.taskNum <= 0:
            self.taskNum = 1
        else:
            self.taskNum = int(math.ceil(self.taskNum))
        self.total_parts = int(self.size / self.partSize)
        if self.total_parts >= 10000:
            self.partSize = int(self.size / 10000) if self.size % 10000 == 0 else int(self.size / 10000) + 1
            self.total_parts = int(self.size / self.partSize)
        if self.size % self.partSize != 0:
            self.total_parts += 1

    def _upload(self):
        if self.headers is None:
            self.headers = UploadFileHeader()

        if self.enableCheckPoint:
            self._load()

        if "uploadId" not in self._record:
            self._prepare()

        unfinished_upload_parts = []
        sent_bytes = const.LONG(0)
        for p in self._record['uploadParts']:
            if not p['isCompleted']:
                unfinished_upload_parts.append(p)
            else:
                sent_bytes += p['length']

        if self.progressCallback is not None:
            self.notifier = progress.ProgressNotifier(self.progressCallback, self.size)
            self.notifier.start()

        try:
            if len(unfinished_upload_parts) > 0:

                if (self.size - sent_bytes) > 0:
                    self.notifier.send(sent_bytes)

                thread_pools = _ThreadPool(functools.partial(self._produce, upload_parts=unfinished_upload_parts),
                                           [self._consume] * self.taskNum)
                thread_pools.run()

                if self._abort:
                    self.obsClient.abortMultipartUpload(self.bucketName, self.objectKey, self._record['uploadId'],
                                                        extensionHeaders=self.extensionHeaders)
                    self.obsClient.log_client.log(
                        ERROR,
                        'the code from server is 4**, please check space, permission and so on.'
                    )
                    self._delete_record()
                    if self._exception is not None:
                        raise Exception(self._exception)

                for p in self._record['uploadParts']:
                    if not p['isCompleted']:
                        if not self.enableCheckPoint:
                            self.obsClient.abortMultipartUpload(self.bucketName, self.objectKey,
                                                                self._record['uploadId'],
                                                                extensionHeaders=self.extensionHeaders)
                        raise Exception('some parts are failed when upload. Please try again')

            part_Etags = []
            for part in self._record['partEtags']:
                part_Etags.append(CompletePart(partNum=part['partNum'], etag=part['etag']))
                self.obsClient.log_client.log(INFO, 'Completing to upload multi_parts')
            resp = self.obsClient.completeMultipartUpload(self.bucketName, self.objectKey, self._record['uploadId'],
                                                          CompleteMultipartUploadRequest(part_Etags),
                                                          extensionHeaders=self.extensionHeaders,
                                                          encoding_type=self.encoding_type)
            self._upload_handle_response(resp)
            return resp
        finally:
            self.notifier.end()

    def _upload_handle_response(self, resp):
        if resp.status < 300:
            if self.enableCheckPoint:
                self._delete_record()
        else:
            if not self.enableCheckPoint:
                self.obsClient.abortMultipartUpload(self.bucketName, self.objectKey, self._record['uploadId'],
                                                    extensionHeaders=self.extensionHeaders)
                self.obsClient.log_client.log(
                    ERROR,
                    'something is wrong when complete multipart.ErrorCode:{0}. ErrorMessage:{1}'.format(
                        resp.errorCode, resp.errorMessage))
            else:
                if 300 < resp.status < 500:
                    self.obsClient.abortMultipartUpload(self.bucketName, self.objectKey, self._record['uploadId'],
                                                        extensionHeaders=self.extensionHeaders)
                    self.obsClient.log_client.log(
                        ERROR,
                        'something is wrong when complete multipart.ErrorCode:{0}. ErrorMessage:{1}'.format(
                            resp.errorCode, resp.errorMessage))
                    self._delete_record()

    def _load(self):
        self._record = self._get_record()
        if self._record and not (self._type_check(self._record) and self._check_upload_record(self._record)):
            if self._record['bucketName'] and self._record['objectKey'] and self._record['uploadId'] is not None:
                self.obsClient.abortMultipartUpload(self._record['bucketName'], self._record['objectKey'],
                                                    self._record['uploadId'], extensionHeaders=self.extensionHeaders)
            self.obsClient.log_client.log(ERROR, 'checkpointFile is invalid')
            self._delete_record()
            self._record = None
        if self._record is None:
            self._record = {'bucketName': self.bucketName, 'objectKey': self.objectKey, 'uploadFile': self.fileName,
                            'partEtags': []}

    def _type_check(self, record):
        try:
            for key in ('bucketName', 'objectKey', 'uploadId', 'uploadFile'):
                if not isinstance(record[key], str):
                    self.obsClient.log_client.log(
                        ERROR,
                        '{0} is not a string type. {1} belong to {2}'.format(key, record[key], record[key].__class__)
                    )
                    return False
            if not isinstance(record['fileStatus'], list):
                self.obsClient.log_client.log(ERROR, 'fileStatus is not a list.It is {0} type'.format(
                    record['fileStatus'].__class__))
                return False
            if not isinstance(record['uploadParts'], list):
                self.obsClient.log_client.log(ERROR, 'uploadParts is not a list.It is {0} type'.format(
                    record['uploadParts'].__class__))
                return False
            if not isinstance(record['partEtags'], list):
                self.obsClient.log_client.log(ERROR, 'partEtags is not a dict.It is {0} type'.format(
                    record['partEtags'].__class__))
                return False
        except KeyError as e:
            self.obsClient.log_client.log(INFO, 'Key is not found:{0}'.format(e.args))
            return False
        return True

    def _check_upload_record(self, record):
        if not ((record['bucketName'] == self.bucketName) and (record['objectKey'] == self.objectKey) and (
                record['uploadFile'] == self.fileName)):
            self.obsClient.log_client.log(INFO,
                                          'the bucketName or objectKey or uploadFile was changed. clear the record')
            return False
        if record['uploadId'] is None:
            self.obsClient.log_client.log(INFO,
                                          '{0} (uploadId) not exist, clear the record.'.format(record['upload_id']))
            return False

        if record['fileStatus'][0] != self.size or record['fileStatus'][1] != self.lastModified:
            self.obsClient.log_client.log(INFO, '{0} was changed, clear the record.'.format(self.fileName))
            return False
        if self.checkSum and len(record['fileStatus']) >= 3:
            checkSum = util.base64_encode(
                util.md5_file_encode_by_size_offset(file_path=self.fileName, size=self.size, offset=0))
            if record['fileStatus'][2] and record['fileStatus'][2] != checkSum:
                self.obsClient.log_client.log(INFO, '{0} content was changed, clear the record.'.format(self.fileName))
                return False
        return True

    def _slice_file(self):
        uploadParts = []

        if self.total_parts == 0:
            part = Part(util.to_long(1), util.to_long(0), util.to_long(0), False)
            uploadParts.append(part)
        else:
            offset = 0
            for i in range(1, self.total_parts + 1, 1):
                part = Part(util.to_long(i), util.to_long(offset), util.to_long(self.partSize), False)
                offset += self.partSize
                uploadParts.append(part)

            if self.size % self.partSize != 0:
                uploadParts[self.total_parts - 1].length = util.to_long(self.size % self.partSize)

        return uploadParts

    def get_init_upload_result(self):
        return self.obsClient.initiateMultipartUpload(self.bucketName, self.objectKey, metadata=self.metadata,
                                                      acl=self.headers.acl,
                                                      storageClass=self.headers.storageClass,
                                                      websiteRedirectLocation=self.headers.websiteRedirectLocation,
                                                      contentType=self.headers.contentType,
                                                      sseHeader=self.headers.sseHeader, expires=self.headers.expires,
                                                      extensionGrants=self.headers.extensionGrants,
                                                      extensionHeaders=self.extensionHeaders,
                                                      encoding_type=self.encoding_type)

    def _prepare(self):
        fileStatus = [self.size, self.lastModified]
        if self.checkSum:
            fileStatus.append(util.base64_encode(util.md5_file_encode_by_size_offset(self.fileName, self.size, 0)))

        resp = self.get_init_upload_result()

        if resp.status > 300:
            raise Exception('initiateMultipartUpload failed. ErrorCode:{0}. ErrorMessage:{1}'.format(resp.errorCode,
                                                                                                     resp.errorMessage))
        uploadId = resp.body.uploadId
        self._record["uploadParts"] = self._slice_file()
        self._record["uploadId"] = uploadId
        self._record["fileStatus"] = fileStatus
        self.obsClient.log_client.log(INFO, 'prepare new upload task success. uploadId = {0}'.format(uploadId))
        if self.enableCheckPoint:
            self._write_record(self._record)

    @staticmethod
    def _produce(ThreadPool, upload_parts):
        for part in upload_parts:
            ThreadPool.put(part)

    def _consume(self, ThreadPool):
        while ThreadPool.ok():
            part = ThreadPool.get()
            if part is None:
                break
            self._upload_part(part)

    def _upload_part(self, part):
        if not self._is_abort():
            try:
                resp = self.get_upload_part_resp(part)
                if resp.status < 300:
                    self._record['uploadParts'][part['partNumber'] - 1]['isCompleted'] = True
                    self._record['partEtags'].append(CompletePart(util.to_int(part['partNumber']), resp.body.etag))
                    if self.enableCheckPoint:
                        with self._lock:
                            self._write_record(self._record)
                else:
                    if 300 < resp.status < 500:
                        self._do_abort('errorCode:{0}, errorMessage:{1}'.format(resp.errorCode, resp.errorMessage))
                    self.obsClient.log_client.log(
                        ERROR,
                        'response from server is something wrong. ErrorCode:{0}, ErrorMessage:{1}'.format(
                            resp.errorCode, resp.errorMessage))
            except Exception as e:
                self.obsClient.log_client.log(DEBUG, 'upload part %s error, %s' % (part['partNumber'], e))
                self.obsClient.log_client.log(ERROR, traceback.format_exc())

    def get_upload_part_resp(self, part):
        return self.obsClient._uploadPartWithNotifier(self.bucketName, self.objectKey, part['partNumber'],
                                                      self._record['uploadId'], self.fileName,
                                                      isFile=True, partSize=part['length'],
                                                      offset=part['offset'], notifier=self.notifier,
                                                      extensionHeaders=self.extensionHeaders,
                                                      sseHeader=self.headers.sseHeader)


class downloadOperation(Operation):
    def __init__(self, bucketName, objectKey, downloadFile, partSize, taskNum, enableCheckPoint, checkPointFile,
                 header, versionId, progressCallback, obsClient, imageProcess=None, notifier=progress.NONE_NOTIFIER,
                 extensionHeaders=None):
        if enableCheckPoint and not checkPointFile:
            checkPointFile = downloadFile + '.download_record'
        super(downloadOperation, self).__init__(bucketName, objectKey, downloadFile, partSize, taskNum,
                                                enableCheckPoint,
                                                checkPointFile, progressCallback, obsClient, notifier)
        self.header = header
        self.versionId = versionId
        self.imageProcess = imageProcess
        self.extensionHeaders = extensionHeaders
        self._record = {'bucketName': self.bucketName, 'objectKey': self.objectKey, 'versionId': self.versionId,
                        'downloadFile': self.fileName, 'imageProcess': self.imageProcess}

        parent_dir = os.path.dirname(self.fileName)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)

        self._tmp_file = self.fileName + '.tmp'
        metadata_resp = self.obsClient.getObjectMetadata(self.bucketName, self.objectKey, self.versionId,
                                                         extensionHeaders=self.extensionHeaders,
                                                         sseHeader=self.header.sseHeader, origin=self.header.origin,
                                                         requestHeaders=self.header.requestHeaders)
        if metadata_resp.status < 300:
            self.lastModified = metadata_resp.body.lastModified
            self.size = metadata_resp.body.contentLength \
                if metadata_resp.body.contentLength is not None and metadata_resp.body.contentLength >= 0 else 0
        else:
            if 400 <= metadata_resp.status < 500:
                self._delete_record()
                self._delete_tmp_file()
            self.obsClient.log_client.log(
                ERROR,
                'there are something wrong when touch the object {0}. ErrorCode:{1}, ErrorMessage:{2}'.format(
                    self.objectKey, metadata_resp.errorCode, metadata_resp.errorMessage))
            raise Exception(
                'there are something wrong when touch the object {0}. ErrorCode:{1}, ErrorMessage:{2}'.format(
                    self.objectKey, metadata_resp.status, metadata_resp.errorMessage))
        self._metadata_resp = metadata_resp

    def _delete_tmp_file(self):
        if os.path.exists(self._tmp_file):
            os.remove(self._tmp_file)

    def _do_rename(self):
        try:
            with open(self.fileName, 'wb') as wf:
                with open(self._tmp_file, 'rb') as rf:
                    while True:
                        chunk = rf.read(const.READ_ONCE_LENGTH)
                        if not chunk:
                            break
                        wf.write(chunk)
            if self.enableCheckPoint:
                self._delete_record()
            self._delete_tmp_file()
            return 1
        except Exception:
            return 0

    def _download(self):
        inner_notifier = False

        if self.enableCheckPoint:
            self._load()

        if "downloadParts" not in self._record:
            self._prepare()

        sent_bytes, unfinished_down_parts = self._download_prepare()

        try:
            if len(unfinished_down_parts) > 0:
                if (self.size - sent_bytes) > 0:
                    if self.progressCallback is not None:
                        self.notifier = progress.ProgressNotifier(self.progressCallback, self.size)
                        inner_notifier = True
                        self.notifier.start()
                    self.notifier.send(sent_bytes)

                thread_pools = _ThreadPool(functools.partial(self._produce, download_parts=unfinished_down_parts),
                                           [self._consume] * self.taskNum)
                thread_pools.run()

                if self._abort:
                    self._delete_record()
                    self._delete_tmp_file()
                    if self._exception is not None:
                        raise Exception(self._exception)

                for p in self._record['downloadParts']:
                    if not p['isCompleted']:
                        raise Exception('some parts are failed when download. Please try again')

            try:
                os.rename(self._tmp_file, self.fileName)
                if self.enableCheckPoint:
                    self._delete_record()
                self.obsClient.log_client.log(INFO, 'download success.')
                return self._metadata_resp
            except Exception as e:
                if self._do_rename():
                    self.obsClient.log_client.log(INFO, 'download success.')
                    return self._metadata_resp
                if not self.enableCheckPoint:
                    self._delete_tmp_file()
                self.obsClient.log_client.log(
                    INFO,
                    'Rename failed. The reason maybe:[the {0} exists, not a file path, not permission]. Please check.')
                raise e
        finally:
            if inner_notifier:
                self.notifier.end()

    def _download_prepare(self):
        sent_bytes = 0
        unfinished_down_parts = []
        for part in self._record['downloadParts']:
            if not part['isCompleted']:
                unfinished_down_parts.append(part)
            else:
                sent_bytes += (part['length'] - part['offset']) + 1
        return sent_bytes, unfinished_down_parts

    def _load(self):
        self._record = self._get_record()
        if self._record is not None and not (
                self._type_record(self._record) and self._check_download_record(self._record)):
            self._delete_record()
            self._delete_tmp_file()
            self._record = None
        if self._record is None:
            self._record = {'bucketName': self.bucketName, 'objectKey': self.objectKey, 'versionId': self.versionId,
                            'downloadFile': self.fileName, 'imageProcess': self.imageProcess}

    def _prepare(self):
        object_status = [self.objectKey, self.size, self.lastModified, self.versionId, self.imageProcess]
        with open(_to_unicode(self._tmp_file), 'wb') as f:
            if self.size > 0:
                f.seek(self.size - 1, 0)
            f.write('b' if const.IS_PYTHON2 else 'b'.encode('UTF-8'))

        tmp_file_status = [os.path.getsize(self._tmp_file), os.path.getmtime(self._tmp_file)]
        self._record["downloadParts"] = self._split_object()
        self._record["objectStatus"] = object_status
        self._record["tmpFileStatus"] = tmp_file_status
        self.obsClient.log_client.log(INFO, 'prepare new download task success.')
        if self.enableCheckPoint:
            self._write_record(self._record)

    def _type_record(self, record):
        try:
            for key in ('bucketName', 'objectKey', 'versionId', 'downloadFile', 'imageProcess'):
                if key == 'versionId' and record['versionId'] is None:
                    continue
                if key == 'imageProcess' and record['imageProcess'] is None:
                    continue
                if not isinstance(record[key], str):
                    self.obsClient.log_client.log(
                        ERROR,
                        '{0} is not a string type. {1} belong to {2}'.format(key, record[key], record[key].__class__)
                    )
                    return False
            if not isinstance(record['downloadParts'], list):
                self.obsClient.log_client.log(ERROR, 'downloadParts is not a list.It is {0} type'.format(
                    record['downloadParts'].__class__))
                return False
            if not isinstance(record['objectStatus'], list):
                self.obsClient.log_client.log(ERROR, 'objectStatus is not a list.It is {0} type'.format(
                    record['objectStatus'].__class__))
                return False
            if not isinstance(record['tmpFileStatus'], list):
                self.obsClient.log_client.log(ERROR, 'tmpFileStatus is not a dict.It is {0} type'.format(
                    record['tmpFileStatus'].__class__))
                return False
        except KeyError as e:
            self.obsClient.log_client.log(INFO, 'Key is not found:{0}'.format(e.args))
            return False
        return True

    def _check_download_record(self, record):
        if not operator.eq([record['bucketName'], record['objectKey'], record['versionId'], record['downloadFile'],
                            record['imageProcess']],
                           [self.bucketName, self.objectKey, self.versionId, self.fileName, self.imageProcess]):
            return False
        if not operator.eq(record['objectStatus'],
                           [self.objectKey, self.size, self.lastModified, self.versionId, self.imageProcess]):
            return False
        if record['tmpFileStatus'][0] != self.size:
            return False
        return True

    def _split_object(self):
        downloadParts = []
        num_counts = int(self.size / self.partSize)
        if num_counts >= 10000:
            self.partSize = int(self.size / 10000) if self.size % 10000 == 0 else int(self.size / 10000) + 1
            num_counts = int(self.size / self.partSize)
        if self.size % self.partSize != 0:
            num_counts += 1
        start = 0
        for i in range(1, num_counts + 1):
            end = start + util.to_long(self.partSize) if i != num_counts else util.to_long(self.size)
            part = Part(util.to_long(i), util.to_long(start), end - 1, False)
            start += self.partSize
            downloadParts.append(part)
        return downloadParts

    @staticmethod
    def _produce(ThreadPool, download_parts):
        for part in download_parts:
            ThreadPool.put(part)

    def _consume(self, ThreadPool):
        while ThreadPool.ok():
            part = ThreadPool.get()
            if part is None:
                break
            self._download_part(part)

    @staticmethod
    def _copy_get_object_header(src_header):
        get_object_header = GetObjectHeader()
        get_object_header.sseHeader = src_header.sseHeader
        get_object_header.if_match = src_header.if_match
        get_object_header.if_none_match = src_header.if_none_match
        get_object_header.if_modified_since = src_header.if_modified_since
        get_object_header.if_unmodified_since = src_header.if_unmodified_since
        return get_object_header

    def _download_part(self, part):
        get_object_request = GetObjectRequest(versionId=self.versionId, imageProcess=self.imageProcess)
        get_object_header = self._copy_get_object_header(self.header)
        get_object_header.range = util.to_string(part['offset']) + '-' + util.to_string(part['length'])
        if not self._is_abort():
            # todo 检视 response 的作用与 part_response 的重构
            response = None
            try:
                resp = self.obsClient.getObject(bucketName=self.bucketName, objectKey=self.objectKey,
                                                getObjectRequest=get_object_request,
                                                headers=get_object_header, notifier=self.notifier,
                                                extensionHeaders=self.extensionHeaders)
                if resp.status < 300:
                    part_response = resp.body.response
                    self._download_part_write(part_response, part)
                    self._record['downloadParts'][part['partNumber'] - 1]['isCompleted'] = True
                    if self.enableCheckPoint:
                        with self._lock:
                            self._write_record(self._record)
                else:
                    if 300 < resp.status < 500:
                        self._do_abort('errorCode:{0}, errorMessage:{1}'.format(resp.errorCode, resp.errorMessage))
                    self._exception += (
                        'response from server is something wrong. ErrorCode:{0}, ErrorMessage:{1}\n'.format(
                            resp.errorCode, resp.errorMessage))
                    self.obsClient.log_client.log(
                        ERROR,
                        'response from server is something wrong. ErrorCode:{0}, ErrorMessage:{1}'.format(
                            resp.errorCode, resp.errorMessage))
            except Exception as e:
                self.obsClient.log_client.log(DEBUG, 'download part %s error, %s' % (part['partNumber'], e))
                self.obsClient.log_client.log(ERROR, traceback.format_exc())
            finally:
                if response is not None:
                    part_response.close()

    def _download_part_write(self, response, part):
        chunk_size = const.READ_ONCE_LENGTH
        if response is not None:
            with open(_to_unicode(self._tmp_file), 'rb+') as fs:
                fs.seek(part['offset'], 0)
                while True:
                    chunk = response.read(chunk_size)
                    if not chunk:
                        break
                    fs.write(chunk)


class Part(BaseModel):
    allowedAttr = {'partNumber': const.LONG, 'offset': const.LONG, 'length': const.LONG, 'isCompleted': bool}

    def __init__(self, partNumber, offset, length, isCompleted=False):
        self.partNumber = partNumber
        self.offset = offset
        self.length = length
        self.isCompleted = isCompleted


def _parse_string(content):
    if const.IS_PYTHON2:
        if isinstance(content, dict):
            return dict([(_parse_string(key), _parse_string(value)) for key, value in content.items()])
        elif isinstance(content, list):
            return [_parse_string(element) for element in content]
        elif isinstance(content, const.UNICODE):
            return content.encode('UTF-8')
    return content


def _to_unicode(data):
    if isinstance(data, bytes):
        return data.decode('UTF-8')
    return data


class _ThreadPool(object):
    def __init__(self, producer, consumers):
        self._producer = producer
        self._consumers = consumers
        self._lock = threading.Lock()
        self._queue = queue.Queue()

        self._threads_consumer = []
        self._threads_producer = []
        self._threading_thread = threading.Thread
        self._exc_info = None
        self._exc_stack = None

    def run(self):
        self._add_and_run(self._threading_thread(target=self._producer_start), self._threads_producer)
        for thread in self._threads_producer:
            thread.join()
        for consumer in self._consumers:
            self._add_and_run(self._threading_thread(target=self._consumer_start, args=(consumer,)),
                              self._threads_consumer)

        for thread in self._threads_consumer:
            thread.join()

        if self._exc_info:
            raise self._exc_info[1]

    def put(self, task):
        if task is not None:
            self._queue.put(task)

    def get(self):
        return self._queue.get()

    def ok(self):
        with self._lock:
            return self._exc_info is None

    @staticmethod
    def _add_and_run(thread, pool):
        thread.daemon = True
        thread.start()
        pool.append(thread)

    def _producer_start(self):
        try:
            self._producer(self)
        except Exception:
            with self._lock:
                if self._exc_info is None:
                    self._exc_info = sys.exc_info()
                    self._exc_stack = traceback.format_exc()
        finally:
            self._put_end()

    def _consumer_start(self, consumer):
        try:
            consumer(self)
        except Exception:
            with self._lock:
                if self._exc_info is None:
                    self._exc_info = sys.exc_info()
                    self._exc_stack = traceback.format_exc()

    def _put_end(self):
        length = len(self._consumers)
        for _ in range(length):
            self._queue.put(None)



#!/usr/bin/python
# -*- coding:utf-8 -*-
# Copyright 2019 Huawei Technologies Co.,Ltd.
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License.  You may obtain a copy of the
# License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.

import os
import sys
import logging.handlers
from obs.const import IS_WINDOWS, IS_PYTHON2

import threading

_lock = threading.Lock()

if IS_PYTHON2:
    from ConfigParser import ConfigParser
else:
    from configparser import ConfigParser

_handler = logging.handlers.RotatingFileHandler

CRITICAL = logging.CRITICAL
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG

LOG_LEVEL_DICT = {'CRITICAL': CRITICAL, 'ERROR': ERROR, 'WARNING': WARNING, 'INFO': INFO, 'DEBUG': DEBUG}


class LogConf(object):
    def __init__(self, config_file=None, sec='LOGCONF'):
        self.disable = True
        if config_file:
            str_path = os.path.abspath(config_file)
            if not os.path.exists(str_path):
                raise Exception('%s is not exist' % (str_path))

            cf = ConfigParser()
            read_ok = cf.read(config_file) if IS_PYTHON2 else cf.read(config_file, 'UTF-8')

            if read_ok:
                secs = cf.sections()
                if sec not in secs:
                    raise Exception('%s is not in secs:%s' % (sec, str(secs)))

                items = cf.items(sec)
                idict = {}
                for e in items:
                    idict[e[0]] = e[1]
                self.log_file_dir = idict.get('logfiledir', './')
                self.log_file_name = idict.get('logfilename', 'log.log')
                self.log_file_number = int(idict.get('logfilenumber', 0))
                self.log_file_size = int(idict.get('logfilesize', 0))
                self.print_log_to_console = int(idict.get('printlogtoconsole', 0))
                log_file_level = idict.get('logfilelevel')
                print_log_level = idict.get('printloglevel')
                self.log_file_level = LOG_LEVEL_DICT.get(log_file_level, DEBUG)
                self.print_log_level = LOG_LEVEL_DICT.get(print_log_level, DEBUG)
                self.disable = False


class NoneLogClient(object):
    def log(self, level, msg, *args, **kwargs):
        pass

    def close(self):
        pass


class LogClient(object):
    def __init__(self, log_config, log_name='OBS_LOGGER', display_name=None):
        if not log_config or not isinstance(log_config, LogConf):
            raise Exception('log config is not correct')
        self.log_config = log_config
        if display_name is None:
            display_name = log_name
        self.display_name = display_name
        self.logger = logging.getLogger(log_name)
        if not hasattr(self.logger, '_inited'):
            self.logger.setLevel(logging.DEBUG)
            if hasattr(self.logger, 'propagate'):
                self.logger.propagate = False
            if not log_config.disable:
                self.initLogger()
            self.logger._inited = 1

    def initLogger(self):
        if not os.path.exists(self.log_config.log_file_dir):
            with _lock:
                if not os.path.exists(self.log_config.log_file_dir):
                    os.makedirs(self.log_config.log_file_dir, 0o755)

        sep = '\\' if IS_WINDOWS else '/'
        logfilepath = self.log_config.log_file_dir + sep + self.log_config.log_file_name
        encoding = None if IS_PYTHON2 else 'UTF-8'
        formatter_handle = _handler(filename=logfilepath, encoding=encoding,
                                    maxBytes=1024 * 1024 * self.log_config.log_file_size,
                                    backupCount=self.log_config.log_file_number)
        formatter_handle.setLevel(self.log_config.log_file_level)
        formatter = logging.Formatter(
            '%(asctime)s|process:%(process)d|thread:%(thread)d|%(levelname)s|HTTP(s)+XML|%(message)s|')
        formatter_handle.setFormatter(formatter)

        self.logger.addHandler(formatter_handle)
        if self.log_config.print_log_to_console == 1:
            console_handler = logging.StreamHandler()
            console_handler.setLevel(self.log_config.print_log_level)
            console_handler.setFormatter(formatter)
            self.logger.addHandler(console_handler)

    def close(self):
        for handle in self.logger.handlers:
            try:
                handle.close()
            except Exception:
                # ignore exception of log handle close.
                pass

    def log(self, level, msg, *args, **kwargs):
        base_back = sys._getframe().f_back
        func_name = base_back.f_code.co_name
        while func_name.lower() == 'log':
            base_back = base_back.f_back
            func_name = base_back.f_code.co_name
        line = base_back.f_lineno
        msg = '%(logger)s|%(name)s,%(lineno)d|' % {'logger': self.display_name, 'name': func_name,
                                                   'lineno': int(line)} + str(msg)

        if level == CRITICAL:
            self.logger.critical(msg, *args, **kwargs)
        elif level == ERROR:
            self.logger.error(msg, *args, **kwargs)
        elif level == WARNING:
            self.logger.warning(msg, *args, **kwargs)
        elif level == INFO:
            self.logger.info(msg, *args, **kwargs)
        elif level == DEBUG:
            self.logger.debug(msg, *args, **kwargs)
